@@ -6,7 +6,7 @@
  *	License as published by the Free Software Foundation version 2.1
  *	of the License.
  *
- * Copyright (c) 2003-2008 Thomas Graf <tgraf@suug.ch>
+ * Copyright (c) 2003-2011 Thomas Graf <tgraf@suug.ch>
  */
 
 #ifndef NETLINK_LOCAL_H_
@@ -28,10 +28,13 @@
 #include <inttypes.h>
 #include <assert.h>
 #include <limits.h>
+#include <search.h>
 
 #include <arpa/inet.h>
 #include <netdb.h>
 
+#include <defs.h>
+
 #ifndef SOL_NETLINK
 #define SOL_NETLINK 270
 #endif
@@ -42,10 +45,19 @@
 #include <linux/if.h>
 #include <linux/if_arp.h>
 #include <linux/if_ether.h>
+#include <linux/ethtool.h>
 #include <linux/pkt_sched.h>
 #include <linux/pkt_cls.h>
 #include <linux/gen_stats.h>
 #include <linux/ip_mp_alg.h>
+#include <linux/atm.h>
+#include <linux/inetdevice.h>
+#include <linux/ipv6.h>
+#include <linux/snmp.h>
+
+#ifndef DISABLE_PTHREADS
+#include <pthread.h>
+#endif
 
 #include <netlink/netlink.h>
 #include <netlink/handlers.h>
@@ -76,45 +88,48 @@ struct trans_list {
 			fprintf(stderr, "DBG<" #LVL ">: " FMT, ##ARG); \
 	} while (0)
 
-#define BUG()                            \
-	do {                                 \
-		fprintf(stderr, "BUG: %s:%d\n",  \
-			__FILE__, __LINE__);         \
-		assert(0);	\
+#define BUG()                            				\
+	do {                                 				\
+		NL_DBG(1, "BUG: %s:%d\n",  				\
+			__FILE__, __LINE__);         			\
+		assert(0);						\
 	} while (0)
 
+#define APPBUG(msg)							\
+	do {								\
+		NL_DBG(1, "APPLICATION BUG: %s:%d:%s: %s\n",		\
+			__FILE__, __LINE__, __PRETTY_FUNCTION__, msg);	\
+		assert(0);						\
+	} while(0)
+
 extern int __nl_read_num_str_file(const char *path,
 				  int (*cb)(long, const char *));
 
 extern int __trans_list_add(int, const char *, struct nl_list_head *);
 extern void __trans_list_clear(struct nl_list_head *);
 
-extern char *__type2str(int, char *, size_t, struct trans_tbl *, size_t);
-extern int __str2type(const char *, struct trans_tbl *, size_t);
+extern char *__type2str(int, char *, size_t, const struct trans_tbl *, size_t);
+extern int __str2type(const char *, const struct trans_tbl *, size_t);
 
 extern char *__list_type2str(int, char *, size_t, struct nl_list_head *);
 extern int __list_str2type(const char *, struct nl_list_head *);
 
-extern char *__flags2str(int, char *, size_t, struct trans_tbl *, size_t);
-extern int __str2flags(const char *, struct trans_tbl *, size_t);
+extern char *__flags2str(int, char *, size_t, const struct trans_tbl *, size_t);
+extern int __str2flags(const char *, const struct trans_tbl *, size_t);
 
 extern void dump_from_ops(struct nl_object *, struct nl_dump_params *);
 
-static inline struct nl_cache *dp_cache(struct nl_object *obj)
-{
-	if (obj->ce_cache == NULL)
-		return nl_cache_mngt_require(obj->ce_ops->oo_name);
-
-	return obj->ce_cache;
-}
-
 static inline int nl_cb_call(struct nl_cb *cb, int type, struct nl_msg *msg)
 {
 	return cb->cb_set[type](msg, cb->cb_args[type]);
 }
 
 #define ARRAY_SIZE(X) (sizeof(X) / sizeof((X)[0]))
+
+/* This is also defined in stddef.h */
+#ifndef offsetof
 #define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
+#endif
 
 #define __init __attribute__ ((constructor))
 #define __exit __attribute__ ((destructor))
@@ -146,8 +161,8 @@ static inline void rtnl_copy_ratespec(struct rtnl_ratespec *dst,
 				      struct tc_ratespec *src)
 {
 	dst->rs_cell_log = src->cell_log;
-	dst->rs_feature = src->feature;
-	dst->rs_addend = src->addend;
+	dst->rs_overhead = src->overhead;
+	dst->rs_cell_align = src->cell_align;
 	dst->rs_mpu = src->mpu;
 	dst->rs_rate = src->rate;
 }
@@ -156,8 +171,8 @@ static inline void rtnl_rcopy_ratespec(struct tc_ratespec *dst,
 				       struct rtnl_ratespec *src)
 {
 	dst->cell_log = src->rs_cell_log;
-	dst->feature = src->rs_feature;
-	dst->addend = src->rs_addend;
+	dst->overhead = src->rs_overhead;
+	dst->cell_align = src->rs_cell_align;
 	dst->mpu = src->rs_mpu;
 	dst->rate = src->rs_rate;
 }
@@ -181,4 +196,62 @@ static inline int wait_for_ack(struct nl_sock *sk)
 		return nl_wait_for_ack(sk);
 }
 
+static inline int build_sysconf_path(char **strp, const char *filename)
+{
+	char *sysconfdir;
+
+	sysconfdir = getenv("NLSYSCONFDIR");
+
+	if (!sysconfdir)
+		sysconfdir = SYSCONFDIR;
+
+	return asprintf(strp, "%s/%s", sysconfdir, filename);
+}
+
+#ifndef DISABLE_PTHREADS
+#define NL_LOCK(NAME) pthread_mutex_t (NAME) = PTHREAD_MUTEX_INITIALIZER
+#define NL_RW_LOCK(NAME) pthread_rwlock_t (NAME) = PTHREAD_RWLOCK_INITIALIZER
+
+static inline void nl_lock(pthread_mutex_t *lock)
+{
+	pthread_mutex_lock(lock);
+}
+
+static inline void nl_unlock(pthread_mutex_t *lock)
+{
+	pthread_mutex_unlock(lock);
+}
+
+static inline void nl_read_lock(pthread_rwlock_t *lock)
+{
+	pthread_rwlock_rdlock(lock);
+}
+
+static inline void nl_read_unlock(pthread_rwlock_t *lock)
+{
+	pthread_rwlock_unlock(lock);
+}
+
+static inline void nl_write_lock(pthread_rwlock_t *lock)
+{
+	pthread_rwlock_wrlock(lock);
+}
+
+static inline void nl_write_unlock(pthread_rwlock_t *lock)
+{
+	pthread_rwlock_unlock(lock);
+}
+
+#else
+#define NL_LOCK(NAME) int __unused_lock_ ##NAME __attribute__((unused))
+#define NL_RW_LOCK(NAME) int __unused_lock_ ##NAME __attribute__((unused))
+
+#define nl_lock(LOCK) do { } while(0)
+#define nl_unlock(LOCK) do { } while(0)
+#define nl_read_lock(LOCK) do { } while(0)
+#define nl_read_unlock(LOCK) do { } while(0)
+#define nl_write_lock(LOCK) do { } while(0)
+#define nl_write_unlock(LOCK) do { } while(0)
+#endif
+
 #endif
